"""A dummy setup file. It may not be required in the future.

See the warning at
https://setuptools.pypa.io/en/latest/setuptools.html
"""
import setuptools


if __name__ == "__main__":
    setuptools.setup()
